/*
 * Copyright (c) 2005,2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef _HARMONICA_H
#define _HARMONICA_H

/* Can Ids of various controller components */
typedef struct CAN_OPEN_PARAM_STR {
	u_int16_t txPDO1;
	u_int16_t txPDO2;
	u_int16_t rxPDO2;
	u_int16_t txSDO;
	u_int16_t rxSDO;
} CAN_OPEN_PARAM_STR;

/* motor state */
typedef enum CAN_MOTOR_STATE {
	ST_PRE_INITIALIZED,
	ST_OPERATION_ENABLED,
	ST_OPERATION_DISABLED,
	ST_MOTOR_FAILURE
} CAN_MOTOR_STATE;

/* Motion type of the controller */
typedef enum CAN_MOTION_TYPE {
	MOTIONTYPE_VELCTRL,
	MOTIONTYPE_POSCTRL
} CAN_MOTION_TYPE;


/* Misc drive parameters */
typedef struct CAN_DRIVE_PARAMS {
	int driveIdent;
	int encIncrPerRevMot;	/* encoder increments per revolution
				   motor shaft */
	double gearRatio;
	double beltRatio;	/* if drive has belt set ratio, else 1 */
	int sign;		/* direction of motion */
	int velMaxEncIncrS;	/* max velocity */
	int accIncrS;		/* max acceleration */
	int decIncrS;		/* max decceleration */
	double posGearRadToPosMotIncr;
	int encOffsetIncr;	/* position in increments of steerwheel
				   at homing position */
	int homingDigIn;	/* specifies which digital input is used
				   for homing signal */
	int homingSpeed;	/* speed to use for homing (incr/s) */
	int homingDelay;	/* delay to allow for homing (s) */
	bool isSteering;	/* true if steering motor,
				   false if drive motor */
	double currToTorque;	/* factor to convert motor active current [A]
				   into torque [Nm] */
	double currMax;		/* max. current allowed */
} CAN_DRIVE_PARAMS;

#define driveParamGetDriveIdent(p) ((p)->driveIdent)
#define driveParamGetSign(p) ((p)->sign)
#define driveParamGetVelMax(p) ((p)->velMaxEncIncrS)

/* Bits from digital inputs */
#define HARMONICA_INPUT_FORWARD_LIMIT 0x0400
#define HARMONICA_INPUT_REVERSE_LIMIT 0x0800

/* Structure describing one particular motor controller */
typedef struct CAN_HARMONICA_STR {
	CAN_DEV *dev;

	const CAN_OPEN_PARAM_STR *canOpen;
	struct CAN_DRIVE_PARAMS driveParam;
	int divForRequestStatus;
	double canTimeout;

	CAN_MOTION_TYPE motionType;
	int statusCtrl;

	bool watchdogActive;

	struct timespec *currentTime;
	struct timespec *watchdogTime;
	struct timespec *velCalcTime;
	struct timespec *failureStartTime;
	struct timespec *sendTime;

	int refVelGearIncS;	/* reference velocity */

	double velGearMeasRadS;
	double posGearMeasRad;
	double prevPosGearMeasRad;

	u_int32_t digIn;
	bool homePos;		/* Reached home position  */

	double oldPos;

	int motorState;
	int newMotorState;

	int countRequestDiv;

	bool currentLimitOn;
	double motorCurr;	/* current (A) */

} CAN_HARMONICA_STR;



/* Error codes */
#define M_harmonicaLib               635
#define S_harmonicaLib_NO_MEMORY     (M_harmonicaLib << 16 | 0)
#define S_harmonicaLib_MOTOR_FAILURE (M_harmonicaLib << 16 | 1)

/* static H2_ERROR[] */
#define CANMOTORS_H2_ERR_MSGS {\
    {"NO_MEMORY", (S_harmonicaLib_NO_MEMORY & 0xffff)},	   \
    {"MOTOR_FAILURE", (S_harmonicaLib_MOTOR_FAILURE & 0xffff)} \
}

extern STATUS intprtSetInt(CAN_HARMONICA_STR *, int, char, char, int, int);
extern CAN_HARMONICA_STR *harmonicaInit(CAN_DEV *, int, CAN_MOTION_TYPE,
    const CAN_OPEN_PARAM_STR *, const CAN_DRIVE_PARAMS *);
extern void harmonicaEnd(CAN_HARMONICA_STR *);
extern STATUS harmonicaStart(CAN_HARMONICA_STR *);
extern void harmonicaStop(CAN_HARMONICA_STR *);
extern STATUS harmonicaInitCtrl(CAN_HARMONICA_STR *);
extern STATUS harmonicaReset(CAN_HARMONICA_STR *);
extern STATUS harmonicaStartWatchdog(CAN_HARMONICA_STR *, bool);
extern void harmonicaRequestDigIn(CAN_HARMONICA_STR *);
extern void harmonicaRequestCurrent(CAN_HARMONICA_STR *);
extern void harmonicaSetGearVelRadS(CAN_HARMONICA_STR *, double);
extern void harmonicaSetGearPosVelRadS(CAN_HARMONICA_STR *, double, double);
extern void harmonicaGetPosRad(CAN_HARMONICA_STR *, double *);
extern void harmonicaGetPosVelRadS(CAN_HARMONICA_STR *, double *, double *);
extern void harmonicaGetDeltaPosVelRadS(CAN_HARMONICA_STR *, 
    double *, double *);
extern void harmonicaGetCurrent(CAN_HARMONICA_STR *, double *);
extern CAN_MOTOR_STATE harmonicaGetMotorState(CAN_HARMONICA_STR *);
extern bool harmonicaDecode(CAN_HARMONICA_STR *, CAN_MSG *);
extern void harmonicaShow(CAN_HARMONICA_STR *);
extern void harmonicaInitHoming(CAN_HARMONICA_STR *);
extern STATUS harmonicaExecHoming(CAN_HARMONICA_STR *);

extern void driveParamInit(CAN_DRIVE_PARAMS *, int);
extern void driveParamPosVelRadToIncr(CAN_DRIVE_PARAMS *, double, double,
    int *, int *);
extern int driveParamPosGearRadToPosMotIncr(CAN_DRIVE_PARAMS *, double);
extern double driveParamPosMotIncrToPosGearRad(CAN_DRIVE_PARAMS *, int);
extern int driveParamVelGearRadSToVelMotIncrPeriod(CAN_DRIVE_PARAMS *, double);
extern double driveParamVelMotIncrPeriodToVelGearRadS(CAN_DRIVE_PARAMS *, int);

#endif
